@@ -2,16 +2,21 @@
 
 use super::*;
 
-use alloc::vec::Vec;
+use alloc::vec::{self, Vec};
+use tinyvec_macros::impl_mirrored;
+
+#[cfg(feature = "serde")]
+use core::marker::PhantomData;
+#[cfg(feature = "serde")]
+use serde::de::{Deserialize, Deserializer, SeqAccess, Visitor};
+#[cfg(feature = "serde")]
+use serde::ser::{Serialize, SerializeSeq, Serializer};
 
 /// Helper to make a `TinyVec`.
 ///
 /// You specify the backing array type, and optionally give all the elements you
 /// want to initially place into the array.
 ///
-/// As an unfortunate restriction, the backing array type must support `Default`
-/// for it to work with this macro.
-///
 /// ```rust
 /// use tinyvec::*;
 ///
@@ -26,6 +31,7 @@ use alloc::vec::Vec;
 /// let many_ints: TinyVec<[i32; 4]> = tiny_vec!(1, 2, 3, 4, 5, 6, 7, 8, 9, 10);
 /// ```
 #[macro_export]
+#[cfg_attr(docs_rs, doc(cfg(feature = "alloc")))]
 macro_rules! tiny_vec {
   ($array_type:ty => $($elem:expr),* $(,)?) => {
     {
@@ -82,13 +88,15 @@ pub enum TinyVecConstructor<A: Array> {
 /// let some_ints = tiny_vec!([i32; 4] => 1, 2, 3);
 /// ```
 #[derive(Clone)]
+#[cfg_attr(docs_rs, doc(cfg(feature = "alloc")))]
 pub enum TinyVec<A: Array> {
   #[allow(missing_docs)]
   Inline(ArrayVec<A>),
   #[allow(missing_docs)]
   Heap(Vec<A::Item>),
 }
-impl<A: Array + Default> Default for TinyVec<A> {
+
+impl<A: Array> Default for TinyVec<A> {
   #[inline]
   #[must_use]
   fn default() -> Self {
@@ -98,24 +106,21 @@ impl<A: Array + Default> Default for TinyVec<A> {
 
 impl<A: Array> Deref for TinyVec<A> {
   type Target = [A::Item];
-  #[inline(always)]
-  #[must_use]
-  fn deref(&self) -> &Self::Target {
-    match self {
-      TinyVec::Inline(a) => a.deref(),
-      TinyVec::Heap(v) => v.deref(),
-    }
+
+  impl_mirrored! {
+    type Mirror = TinyVec;
+    #[inline(always)]
+    #[must_use]
+    fn deref(self: &Self) -> &Self::Target;
   }
 }
 
 impl<A: Array> DerefMut for TinyVec<A> {
-  #[inline(always)]
-  #[must_use]
-  fn deref_mut(&mut self) -> &mut Self::Target {
-    match self {
-      TinyVec::Inline(a) => a.deref_mut(),
-      TinyVec::Heap(v) => v.deref_mut(),
-    }
+  impl_mirrored! {
+    type Mirror = TinyVec;
+    #[inline(always)]
+    #[must_use]
+    fn deref_mut(self: &mut Self) -> &mut Self::Target;
   }
 }
 
@@ -136,8 +141,43 @@ impl<A: Array, I: SliceIndex<[A::Item]>> IndexMut<I> for TinyVec<A> {
   }
 }
 
+#[cfg(feature = "serde")]
+#[cfg_attr(docs_rs, doc(cfg(feature = "serde")))]
+impl<A: Array> Serialize for TinyVec<A>
+where
+  A::Item: Serialize,
+{
+  #[must_use]
+  fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+  where
+    S: Serializer,
+  {
+    let mut seq = serializer.serialize_seq(Some(self.len()))?;
+    for element in self.iter() {
+      seq.serialize_element(element)?;
+    }
+    seq.end()
+  }
+}
+
+#[cfg(feature = "serde")]
+#[cfg_attr(docs_rs, doc(cfg(feature = "serde")))]
+impl<'de, A: Array> Deserialize<'de> for TinyVec<A>
+where
+  A::Item: Deserialize<'de>,
+{
+  fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
+  where
+    D: Deserializer<'de>,
+  {
+    deserializer.deserialize_seq(TinyVecVisitor(PhantomData))
+  }
+}
+
 impl<A: Array> TinyVec<A> {
   /// Returns whether elements are on heap
+  #[inline(always)]
+  #[must_use]
   pub fn is_heap(&self) -> bool {
     match self {
       TinyVec::Heap(_) => true,
@@ -145,6 +185,8 @@ impl<A: Array> TinyVec<A> {
     }
   }
   /// Returns whether elements are on stack
+  #[inline(always)]
+  #[must_use]
   pub fn is_inline(&self) -> bool {
     !self.is_heap()
   }
@@ -160,10 +202,7 @@ impl<A: Array> TinyVec<A> {
   /// tv.shrink_to_fit();
   /// assert!(tv.is_inline());
   /// ```
-  pub fn shrink_to_fit(&mut self)
-  where
-    A: Default,
-  {
+  pub fn shrink_to_fit(&mut self) {
     let vec = match self {
       TinyVec::Inline(_) => return,
       TinyVec::Heap(h) => h,
@@ -173,10 +212,10 @@ impl<A: Array> TinyVec<A> {
       return vec.shrink_to_fit();
     }
 
-    let movedvec = core::mem::replace(vec, Vec::new());
+    let moved_vec = core::mem::replace(vec, Vec::new());
 
     let mut av = ArrayVec::default();
-    let mut rest = av.fill(movedvec);
+    let mut rest = av.fill(moved_vec);
     debug_assert!(rest.next().is_none());
     *self = TinyVec::Inline(av);
   }
@@ -252,7 +291,7 @@ impl<A: Array> TinyVec<A> {
   /// ```text
   /// Note that the allocator may give the collection more space than it requests.
   /// Therefore, capacity can not be relied upon to be precisely minimal.
-  /// Prefer reserve if future insertions are expected.
+  /// Prefer `reserve` if future insertions are expected.
   /// ```
   /// ```rust
   /// use tinyvec::*;
@@ -276,6 +315,30 @@ impl<A: Array> TinyVec<A> {
     /* In this place array has enough place, so no work is needed more */
     return;
   }
+
+  /// Makes a new TinyVec with _at least_ the given capacity.
+  ///
+  /// If the requested capacity is less than or equal to the array capacity you
+  /// get an inline vec. If it's greater than you get a heap vec.
+  /// ```
+  /// # use tinyvec::*;
+  /// let t = TinyVec::<[u8; 10]>::with_capacity(5);
+  /// assert!(t.is_inline());
+  /// assert!(t.capacity() >= 5);
+  ///
+  /// let t = TinyVec::<[u8; 10]>::with_capacity(20);
+  /// assert!(t.is_heap());
+  /// assert!(t.capacity() >= 20);
+  /// ```
+  #[inline]
+  #[must_use]
+  pub fn with_capacity(cap: usize) -> Self {
+    if cap <= A::CAPACITY {
+      TinyVec::Inline(ArrayVec::default())
+    } else {
+      TinyVec::Heap(Vec::with_capacity(cap))
+    }
+  }
 }
 
 impl<A: Array> TinyVec<A> {
@@ -303,62 +366,125 @@ impl<A: Array> TinyVec<A> {
     }
   }
 
-  /// A mutable pointer to the backing array.
+  impl_mirrored! {
+    type Mirror = TinyVec;
+
+    /// Remove an element, swapping the end of the vec into its place.
+    ///
+    /// ## Panics
+    /// * If the index is out of bounds.
+    ///
+    /// ## Example
+    /// ```rust
+    /// use tinyvec::*;
+    /// let mut tv = tiny_vec!([&str; 4] => "foo", "bar", "quack", "zap");
+    ///
+    /// assert_eq!(tv.swap_remove(1), "bar");
+    /// assert_eq!(tv.as_slice(), &["foo", "zap", "quack"][..]);
+    ///
+    /// assert_eq!(tv.swap_remove(0), "foo");
+    /// assert_eq!(tv.as_slice(), &["quack", "zap"][..]);
+    /// ```
+    #[inline]
+    pub fn swap_remove(self: &mut Self, index: usize) -> A::Item;
+
+    /// Remove and return the last element of the vec, if there is one.
+    ///
+    /// ## Failure
+    /// * If the vec is empty you get `None`.
+    #[inline]
+    pub fn pop(self: &mut Self) -> Option<A::Item>;
+
+    /// Removes the item at `index`, shifting all others down by one index.
+    ///
+    /// Returns the removed element.
+    ///
+    /// ## Panics
+    ///
+    /// If the index is out of bounds.
+    ///
+    /// ## Example
+    ///
+    /// ```rust
+    /// use tinyvec::*;
+    /// let mut tv = tiny_vec!([i32; 4] => 1, 2, 3);
+    /// assert_eq!(tv.remove(1), 2);
+    /// assert_eq!(tv.as_slice(), &[1, 3][..]);
+    /// ```
+    #[inline]
+    pub fn remove(self: &mut Self, index: usize) -> A::Item;
+
+    /// The length of the vec (in elements).
+    #[inline(always)]
+    #[must_use]
+    pub fn len(self: &Self) -> usize;
+
+    /// The capacity of the `TinyVec`.
+    ///
+    /// When not heap allocated this is fixed based on the array type.
+    /// Otherwise its the result of the underlying Vec::capacity.
+    #[inline(always)]
+    #[must_use]
+    pub fn capacity(self: &Self) -> usize;
+
+    /// Reduces the vec's length to the given value.
+    ///
+    /// If the vec is already shorter than the input, nothing happens.
+    #[inline]
+    pub fn truncate(self: &mut Self, new_len: usize);
+
+    /// A mutable pointer to the backing array.
+    ///
+    /// ## Safety
+    ///
+    /// This pointer has provenance over the _entire_ backing array/buffer.
+    #[inline(always)]
+    #[must_use]
+    pub fn as_mut_ptr(self: &mut Self) -> *mut A::Item;
+
+    /// A const pointer to the backing array.
+    ///
+    /// ## Safety
+    ///
+    /// This pointer has provenance over the _entire_ backing array/buffer.
+    #[inline(always)]
+    #[must_use]
+    pub fn as_ptr(self: &Self) -> *const A::Item;
+  }
+
+  /// Walk the vec and keep only the elements that pass the predicate given.
   ///
-  /// ## Safety
+  /// ## Example
   ///
-  /// This pointer has provenance over the _entire_ backing array/buffer.
-  #[inline(always)]
-  #[must_use]
-  pub fn as_mut_ptr(&mut self) -> *mut A::Item {
+  /// ```rust
+  /// use tinyvec::*;
+  ///
+  /// let mut tv = tiny_vec!([i32; 10] => 1, 2, 3, 4);
+  /// tv.retain(|&x| x % 2 == 0);
+  /// assert_eq!(tv.as_slice(), &[2, 4][..]);
+  /// ```
+  #[inline]
+  pub fn retain<F: FnMut(&A::Item) -> bool>(self: &mut Self, acceptable: F) {
     match self {
-      TinyVec::Inline(a) => a.as_mut_ptr(),
-      TinyVec::Heap(v) => v.as_mut_ptr(),
+      TinyVec::Inline(i) => i.retain(acceptable),
+      TinyVec::Heap(h) => h.retain(acceptable),
     }
   }
 
   /// Helper for getting the mut slice.
   #[inline(always)]
   #[must_use]
-  pub fn as_mut_slice(&mut self) -> &mut [A::Item] {
+  pub fn as_mut_slice(self: &mut Self) -> &mut [A::Item] {
     self.deref_mut()
   }
 
-  /// A const pointer to the backing array.
-  ///
-  /// ## Safety
-  ///
-  /// This pointer has provenance over the _entire_ backing array/buffer.
-  #[inline(always)]
-  #[must_use]
-  pub fn as_ptr(&self) -> *const A::Item {
-    match self {
-      TinyVec::Inline(a) => a.as_ptr(),
-      TinyVec::Heap(v) => v.as_ptr(),
-    }
-  }
-
   /// Helper for getting the shared slice.
   #[inline(always)]
   #[must_use]
-  pub fn as_slice(&self) -> &[A::Item] {
+  pub fn as_slice(self: &Self) -> &[A::Item] {
     self.deref()
   }
 
-  /// The capacity of the `TinyVec`.
-  ///
-  /// When not heap allocated this is fixed based on the array type.
-  /// Otherwise its the result of the underlying Vec::capacity.
-  #[inline(always)]
-  #[must_use]
-  pub fn capacity(&self) -> usize {
-    match self {
-      // Note: this shouldn't use A::CAPACITY. See ArrayVec::capacity().
-      TinyVec::Inline(v) => v.capacity(),
-      TinyVec::Heap(v) => v.capacity(),
-    }
-  }
-
   /// Removes all elements from the vec.
   #[inline(always)]
   pub fn clear(&mut self) {
@@ -403,6 +529,11 @@ impl<A: Array> TinyVec<A> {
   /// Creates a draining iterator that removes the specified range in the vector
   /// and yields the removed items.
   ///
+  /// **Note: This method has significant performance issues compared to
+  /// matching on the TinyVec and then calling drain on the Inline or Heap value
+  /// inside. The draining iterator has to branch on every single access. It is
+  /// provided for simplicity and compatability only.**
+  ///
   /// ## Panics
   /// * If the start is greater than the end
   /// * If the end is past the edge of the vec.
@@ -423,33 +554,9 @@ impl<A: Array> TinyVec<A> {
     &mut self,
     range: R,
   ) -> TinyVecDrain<'_, A> {
-    use core::ops::Bound;
-    let start = match range.start_bound() {
-      Bound::Included(x) => *x,
-      Bound::Excluded(x) => x + 1,
-      Bound::Unbounded => 0,
-    };
-    let end = match range.end_bound() {
-      Bound::Included(x) => x + 1,
-      Bound::Excluded(x) => *x,
-      Bound::Unbounded => self.len(),
-    };
-    assert!(
-      start <= end,
-      "TinyVec::drain> Illegal range, {} to {}",
-      start,
-      end
-    );
-    assert!(
-      end <= self.len(),
-      "TinyVec::drain> Range ends at {} but length is only {}!",
-      end,
-      self.len()
-    );
-    TinyVecDrain {
-      parent: self,
-      target_index: start,
-      target_count: end - start,
+    match self {
+      TinyVec::Inline(i) => TinyVecDrain::Inline(i.drain(range)),
+      TinyVec::Heap(h) => TinyVecDrain::Heap(h.drain(range)),
     }
   }
 
@@ -491,8 +598,11 @@ impl<A: Array> TinyVec<A> {
     }
   }
 
+  /// This is an internal implementation detail of the `tiny_vec!` macro, and
+  /// using it other than from that macro is not supported by this crate's
+  /// SemVer guarantee.
   #[inline(always)]
-  #[doc(hidden)] // Internal implementation details of `tiny_vec!`
+  #[doc(hidden)]
   pub fn constructor_for_capacity(cap: usize) -> TinyVecConstructor<A> {
     if cap <= A::CAPACITY {
       TinyVecConstructor::Inline(TinyVec::Inline)
@@ -548,38 +658,13 @@ impl<A: Array> TinyVec<A> {
     self.len() == 0
   }
 
-  /// The length of the vec (in elements).
-  #[inline(always)]
-  #[must_use]
-  pub fn len(&self) -> usize {
-    match self {
-      TinyVec::Inline(a) => a.len(),
-      TinyVec::Heap(v) => v.len(),
-    }
-  }
-
   /// Makes a new, empty vec.
   #[inline(always)]
   #[must_use]
-  pub fn new() -> Self
-  where
-    A: Default,
-  {
+  pub fn new() -> Self {
     Self::default()
   }
 
-  /// Remove and return the last element of the vec, if there is one.
-  ///
-  /// ## Failure
-  /// * If the vec is empty you get `None`.
-  #[inline]
-  pub fn pop(&mut self) -> Option<A::Item> {
-    match self {
-      TinyVec::Inline(a) => a.pop(),
-      TinyVec::Heap(v) => v.pop(),
-    }
-  }
-
   /// Place an element onto the end of the vec.
   /// ## Panics
   /// * If the length of the vec would overflow the capacity.
@@ -604,30 +689,6 @@ impl<A: Array> TinyVec<A> {
     }
   }
 
-  /// Removes the item at `index`, shifting all others down by one index.
-  ///
-  /// Returns the removed element.
-  ///
-  /// ## Panics
-  ///
-  /// If the index is out of bounds.
-  ///
-  /// ## Example
-  ///
-  /// ```rust
-  /// use tinyvec::*;
-  /// let mut tv = tiny_vec!([i32; 4] => 1, 2, 3);
-  /// assert_eq!(tv.remove(1), 2);
-  /// assert_eq!(tv.as_slice(), &[1, 3][..]);
-  /// ```
-  #[inline]
-  pub fn remove(&mut self, index: usize) -> A::Item {
-    match self {
-      TinyVec::Inline(a) => a.remove(index),
-      TinyVec::Heap(v) => v.remove(index),
-    }
-  }
-
   /// Resize the vec to the new length.
   ///
   /// If it needs to be longer, it's filled with clones of the provided value.
@@ -689,25 +750,6 @@ impl<A: Array> TinyVec<A> {
     }
   }
 
-  /// Walk the vec and keep only the elements that pass the predicate given.
-  ///
-  /// ## Example
-  ///
-  /// ```rust
-  /// use tinyvec::*;
-  ///
-  /// let mut tv = tiny_vec!([i32; 10] => 1, 2, 3, 4);
-  /// tv.retain(|&x| x % 2 == 0);
-  /// assert_eq!(tv.as_slice(), &[2, 4][..]);
-  /// ```
-  #[inline]
-  pub fn retain<F: FnMut(&A::Item) -> bool>(&mut self, acceptable: F) {
-    match self {
-      TinyVec::Inline(a) => a.retain(acceptable),
-      TinyVec::Heap(v) => v.retain(acceptable),
-    }
-  }
-
   /// Splits the collection at the point given.
   ///
   /// * `[0, at)` stays in this vec
@@ -726,10 +768,7 @@ impl<A: Array> TinyVec<A> {
   /// assert_eq!(tv2.as_slice(), &[2, 3][..]);
   /// ```
   #[inline]
-  pub fn split_off(&mut self, at: usize) -> Self
-  where
-    A: Default,
-  {
+  pub fn split_off(&mut self, at: usize) -> Self {
     match self {
       TinyVec::Inline(a) => TinyVec::Inline(a.split_off(at)),
       TinyVec::Heap(v) => TinyVec::Heap(v.split_off(at)),
@@ -772,11 +811,11 @@ impl<A: Array> TinyVec<A> {
     use core::ops::Bound;
     let start = match range.start_bound() {
       Bound::Included(x) => *x,
-      Bound::Excluded(x) => x + 1,
+      Bound::Excluded(x) => x.saturating_add(1),
       Bound::Unbounded => 0,
     };
     let end = match range.end_bound() {
-      Bound::Included(x) => x + 1,
+      Bound::Included(x) => x.saturating_add(1),
       Bound::Excluded(x) => *x,
       Bound::Unbounded => self.len(),
     };
@@ -801,41 +840,6 @@ impl<A: Array> TinyVec<A> {
     }
   }
 
-  /// Remove an element, swapping the end of the vec into its place.
-  ///
-  /// ## Panics
-  /// * If the index is out of bounds.
-  ///
-  /// ## Example
-  /// ```rust
-  /// use tinyvec::*;
-  /// let mut tv = tiny_vec!([&str; 4] => "foo", "bar", "quack", "zap");
-  ///
-  /// assert_eq!(tv.swap_remove(1), "bar");
-  /// assert_eq!(tv.as_slice(), &["foo", "zap", "quack"][..]);
-  ///
-  /// assert_eq!(tv.swap_remove(0), "foo");
-  /// assert_eq!(tv.as_slice(), &["quack", "zap"][..]);
-  /// ```
-  #[inline]
-  pub fn swap_remove(&mut self, index: usize) -> A::Item {
-    match self {
-      TinyVec::Inline(a) => a.swap_remove(index),
-      TinyVec::Heap(v) => v.swap_remove(index),
-    }
-  }
-
-  /// Reduces the vec's length to the given value.
-  ///
-  /// If the vec is already shorter than the input, nothing happens.
-  #[inline]
-  pub fn truncate(&mut self, new_len: usize) {
-    match self {
-      TinyVec::Inline(a) => a.truncate(new_len),
-      TinyVec::Heap(v) => v.truncate(new_len),
-    }
-  }
-
   /// Wraps an array, using the given length as the starting length.
   ///
   /// If you want to use the whole length of the array, you can just use the
@@ -855,29 +859,50 @@ impl<A: Array> TinyVec<A> {
 /// Draining iterator for `TinyVecDrain`
 ///
 /// See [`TinyVecDrain::drain`](TinyVecDrain::<A>::drain)
-pub struct TinyVecDrain<'p, A: Array> {
-  parent: &'p mut TinyVec<A>,
-  target_index: usize,
-  target_count: usize,
+pub enum TinyVecDrain<'p, A: Array> {
+  #[allow(missing_docs)]
+  Inline(ArrayVecDrain<'p, A::Item>),
+  #[allow(missing_docs)]
+  Heap(vec::Drain<'p, A::Item>),
 }
-impl<'p, A: Array> FusedIterator for TinyVecDrain<'p, A> {}
+
 impl<'p, A: Array> Iterator for TinyVecDrain<'p, A> {
   type Item = A::Item;
+
+  impl_mirrored! {
+    type Mirror = TinyVecDrain;
+
+    #[inline]
+    fn next(self: &mut Self) -> Option<Self::Item>;
+    #[inline]
+    fn nth(self: &mut Self, n: usize) -> Option<Self::Item>;
+    #[inline]
+    fn size_hint(self: &Self) -> (usize, Option<usize>);
+    #[inline]
+    fn last(self: Self) -> Option<Self::Item>;
+    #[inline]
+    fn count(self: Self) -> usize;
+  }
+
   #[inline]
-  fn next(&mut self) -> Option<Self::Item> {
-    if self.target_count > 0 {
-      let out = self.parent.remove(self.target_index);
-      self.target_count -= 1;
-      Some(out)
-    } else {
-      None
+  fn for_each<F: FnMut(Self::Item)>(self, f: F) {
+    match self {
+      TinyVecDrain::Inline(i) => i.for_each(f),
+      TinyVecDrain::Heap(h) => h.for_each(f),
     }
   }
 }
-impl<'p, A: Array> Drop for TinyVecDrain<'p, A> {
-  #[inline]
-  fn drop(&mut self) {
-    for _ in self {}
+
+impl<'p, A: Array> DoubleEndedIterator for TinyVecDrain<'p, A> {
+  impl_mirrored! {
+    type Mirror = TinyVecDrain;
+
+    #[inline]
+    fn next_back(self: &mut Self) -> Option<Self::Item>;
+
+    #[cfg(feature = "rustc_1_40")]
+    #[inline]
+    fn nth_back(self: &mut Self, n: usize) -> Option<Self::Item>;
   }
 }
 
@@ -897,6 +922,7 @@ where
 {
   type Item = A::Item;
 
+  #[inline]
   fn next(&mut self) -> Option<A::Item> {
     if self.removal_start < self.removal_end {
       match self.replacement.next() {
@@ -949,6 +975,7 @@ where
   A: Array,
   I: Iterator<Item = A::Item> + DoubleEndedIterator,
 {
+  #[inline]
   fn next_back(&mut self) -> Option<A::Item> {
     if self.removal_start < self.removal_end {
       match self.replacement.next_back() {
@@ -1064,7 +1091,7 @@ impl<A: Array> From<A> for TinyVec<A> {
 impl<T, A> From<&'_ [T]> for TinyVec<A>
 where
   T: Clone + Default,
-  A: Array<Item = T> + Default,
+  A: Array<Item = T>,
 {
   #[inline]
   #[must_use]
@@ -1083,7 +1110,7 @@ where
 impl<T, A> From<&'_ mut [T]> for TinyVec<A>
 where
   T: Clone + Default,
-  A: Array<Item = T> + Default,
+  A: Array<Item = T>,
 {
   #[inline]
   #[must_use]
@@ -1092,7 +1119,7 @@ where
   }
 }
 
-impl<A: Array + Default> FromIterator<A::Item> for TinyVec<A> {
+impl<A: Array> FromIterator<A::Item> for TinyVec<A> {
   #[inline]
   #[must_use]
   fn from_iter<T: IntoIterator<Item = A::Item>>(iter: T) -> Self {
@@ -1111,54 +1138,38 @@ pub enum TinyVecIterator<A: Array> {
 }
 
 impl<A: Array> TinyVecIterator<A> {
-  /// Returns the remaining items of this iterator as a slice.
-  #[inline]
-  #[must_use]
-  pub fn as_slice(&self) -> &[A::Item] {
-    match self {
-      TinyVecIterator::Inline(a) => a.as_slice(),
-      TinyVecIterator::Heap(v) => v.as_slice(),
-    }
+  impl_mirrored! {
+    type Mirror = TinyVecIterator;
+    /// Returns the remaining items of this iterator as a slice.
+    #[inline]
+    #[must_use]
+    pub fn as_slice(self: &Self) -> &[A::Item];
   }
 }
+
 impl<A: Array> FusedIterator for TinyVecIterator<A> {}
+
 impl<A: Array> Iterator for TinyVecIterator<A> {
   type Item = A::Item;
-  #[inline]
-  fn next(&mut self) -> Option<Self::Item> {
-    match self {
-      TinyVecIterator::Inline(a) => a.next(),
-      TinyVecIterator::Heap(v) => v.next(),
-    }
-  }
-  #[inline(always)]
-  #[must_use]
-  fn size_hint(&self) -> (usize, Option<usize>) {
-    match self {
-      TinyVecIterator::Inline(a) => a.size_hint(),
-      TinyVecIterator::Heap(v) => v.size_hint(),
-    }
-  }
-  #[inline(always)]
-  fn count(self) -> usize {
-    match self {
-      TinyVecIterator::Inline(a) => a.count(),
-      TinyVecIterator::Heap(v) => v.count(),
-    }
-  }
-  #[inline]
-  fn last(self) -> Option<Self::Item> {
-    match self {
-      TinyVecIterator::Inline(a) => a.last(),
-      TinyVecIterator::Heap(v) => v.last(),
-    }
-  }
-  #[inline]
-  fn nth(&mut self, n: usize) -> Option<A::Item> {
-    match self {
-      TinyVecIterator::Inline(a) => a.nth(n),
-      TinyVecIterator::Heap(v) => v.nth(n),
-    }
+
+  impl_mirrored! {
+    type Mirror = TinyVecIterator;
+
+    #[inline]
+    fn next(self: &mut Self) -> Option<Self::Item>;
+
+    #[inline(always)]
+    #[must_use]
+    fn size_hint(self: &Self) -> (usize, Option<usize>);
+
+    #[inline(always)]
+    fn count(self: Self) -> usize;
+
+    #[inline]
+    fn last(self: Self) -> Option<Self::Item>;
+
+    #[inline]
+    fn nth(self: &mut Self, n: usize) -> Option<A::Item>;
   }
 }
 
@@ -1426,3 +1437,37 @@ where
     write!(f, "]")
   }
 }
+
+#[cfg(feature = "serde")]
+struct TinyVecVisitor<A: Array>(PhantomData<A>);
+
+#[cfg(feature = "serde")]
+impl<'de, A: Array> Visitor<'de> for TinyVecVisitor<A>
+where
+  A::Item: Deserialize<'de>,
+{
+  type Value = TinyVec<A>;
+
+  fn expecting(
+    &self,
+    formatter: &mut core::fmt::Formatter,
+  ) -> core::fmt::Result {
+    formatter.write_str("a sequence")
+  }
+
+  fn visit_seq<S>(self, mut seq: S) -> Result<Self::Value, S::Error>
+  where
+    S: SeqAccess<'de>,
+  {
+    let mut new_tinyvec = match seq.size_hint() {
+      Some(expected_size) => TinyVec::with_capacity(expected_size),
+      None => Default::default(),
+    };
+
+    while let Some(value) = seq.next_element()? {
+      new_tinyvec.push(value);
+    }
+
+    Ok(new_tinyvec)
+  }
+}
